/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the QtGui module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** No Commercial Usage
** This file contains pre-release code and may not be distributed.
** You may use this file in accordance with the terms and conditions
** contained in the Technology Preview License Agreement accompanying
** this package.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** If you have questions regarding the use of this file, please contact
** Nokia at qt-info@nokia.com.
**
**
**
**
**
**
**
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QLayout>
#include <QAction>
#include <QWidgetAction>
#include <QPushButton>
#include <QApplication>
#include <QDesktopWidget>
#include <QVarLengthArray>

//#include "qx11info_x11.h"

#include <private/qmaemo5style_p.h>
#include <private/qgtkpainter_p.h>
#include <private/qmenubar_p.h>
#include <private/qmenu_maemo5_p.h>
#include <private/qt_x11_p.h>

QT_BEGIN_NAMESPACE

#if defined(Q_WS_MAEMO_5)

void QMenuBarPrivate::maemoResetApplicationMenuAtom(QWidget *w)
{
    if (!w)
        return;

    QWidget *window = w->window();
    bool has_menubar = (window->findChild<QMenuBar *>());

    Atom *oldAtoms = 0;
    int oldCount = 0;

    if (!XGetWMProtocols(QX11Info::display(), window->winId(), &oldAtoms, &oldCount))
        return;

    Atom customContext = XInternAtom(QX11Info::display(), "_NET_WM_CONTEXT_CUSTOM", false);

    // re-create the Atom list with or without the customContext Atom
    QVarLengthArray<Atom> newAtoms(oldCount + 1);
    int newCount = 0;
    bool found = false;

    for (int i = 0; i < oldCount; ++i) {
        if (oldAtoms[i] == customContext) {
            found = true;
            if (!has_menubar) // no menubar anymore -> remove atom
                continue;
        }
        newAtoms[newCount++] = oldAtoms[i];
    }
    if (!found && has_menubar) // first menubar in window -> add atom
        newAtoms[newCount++] = customContext;

    XFree(oldAtoms);
    XSetWMProtocols(QX11Info::display(), window->winId(), newAtoms.data(), newCount);

    // also show the indicator (without this, the menu would work, but the
    // little indicator arrow in the titlebar would be missing)
    if (has_menubar) {
        int on = 1;
        XChangeProperty(X11->display, window->winId(), ATOM(_HILDON_WM_WINDOW_MENU_INDICATOR), XA_INTEGER, 32,
                        PropModeReplace, (unsigned char *) &on, 1);
    } else {
        XDeleteProperty(X11->display, window->winId(), ATOM(_HILDON_WM_WINDOW_MENU_INDICATOR));
    }
}


QMaemo5ApplicationMenu::QMaemo5ApplicationMenu(QMenuBar *menubar)
    : QDialog(menubar->window(), Qt::Dialog | Qt::CustomizeWindowHint), m_selected(0)
{
    setAttribute(Qt::WA_Maemo5ApplicationMenu); // this ends up setting a special XAtom for the Maemo WM

    // just close the menu in case the orientation changes
    connect(QApplication::desktop(), SIGNAL(resized(int)), this, SLOT(close()));

    bool portrait = QApplication::desktop()->availableGeometry().width() < QApplication::desktop()->availableGeometry().height();
    int desktop_width = QApplication::desktop()->availableGeometry().width();
    int maxcol = portrait ? 1 : 2;

    QGridLayout *grid = new QGridLayout(this);
    GtkWidget *hildonAppMenu = QMaemo5StylePrivate::gtkWidget(QLS("HildonAppMenu"));
    guint horizontal_spacing = 16, vertical_spacing = 16, inner_border = 16, external_border = 1;

    if (hildonAppMenu) {
        QMaemo5StylePrivate::gtk_widget_style_get(hildonAppMenu, "horizontal-spacing", &horizontal_spacing, NULL);
        QMaemo5StylePrivate::gtk_widget_style_get(hildonAppMenu, "vertical-spacing", &vertical_spacing, NULL);
        QMaemo5StylePrivate::gtk_widget_style_get(hildonAppMenu, "inner-border", &inner_border, NULL);
        QMaemo5StylePrivate::gtk_widget_style_get(hildonAppMenu, "external-border", &external_border, NULL);
    }

    grid->setHorizontalSpacing(horizontal_spacing);
    grid->setVerticalSpacing(vertical_spacing);
    grid->setContentsMargins(inner_border, inner_border, inner_border, inner_border);
    setFixedWidth(desktop_width - (portrait ? 0 : 2) * external_border);
    int row = 0, col = 0;

    recursiveAddActions(menubar->actions(), grid, row, col, maxcol, QString());
}

void QMaemo5ApplicationMenu::recursiveAddActions(const QList<QAction *> &actions, QGridLayout *grid, int row, int col, int maxcol, const QString &prefix)
{
    foreach (QAction *a, actions) {
        if (!a->isVisible() || !a->isEnabled() || a->isSeparator() || ::qobject_cast<QWidgetAction *>(a)) {
            // we won't show actions that are hidden or disabled, separators and widget actions
            continue;
        } else if (a->menu()) {
            // menu actions will be "flatened" recusively
            QString menu_prefix = prefix;
            if (!menu_prefix.isEmpty())
                menu_prefix.append(QLatin1String(" / "));
            menu_prefix.append(a->text());

            recursiveAddActions(a->menu()->actions(), grid, row, col, maxcol, menu_prefix);
        } else {
            QPushButton *p = new QPushButton(this);
            // we need to add the menu "path" to the action
            QString text = prefix.isEmpty() ? a->text() : prefix + QLatin1String(" / ") + a->text();
            p->setText(text);
            if (a->isCheckable())
                p->setCheckable(true);
            if (a->isChecked())
                p->setChecked(true);
            //if (a->isIconVisibleInMenu())
            //    p->setIcon(a->icon()); // ### should we really do that?

            m_actions.insert(p, a);
            connect(p, SIGNAL(clicked()), this, SLOT(buttonClicked()));
            grid->addWidget(p, row, col++);

            if (col >= maxcol) {
                col = 0;
                ++row;
            }
        }
    }
}

bool QMaemo5ApplicationMenu::isEmpty() const
{
    return m_actions.isEmpty();
}

void QMaemo5ApplicationMenu::paintEvent(QPaintEvent *)
{
    QPainter painter(this);

    GtkWidget *hildonAppMenu = QMaemo5StylePrivate::gtkWidget(QLS("HildonAppMenu"));
    if (hildonAppMenu) {
        QGtkPainter gtkPainter(&painter);
        gtkPainter.setAlphaSupport(false); // Note, alpha disabled for performance reasons
        gtkPainter.paintFlatBox(hildonAppMenu, NULL, rect(), GTK_STATE_NORMAL, GTK_SHADOW_NONE, hildonAppMenu->style, QString());
    }
}

void QMaemo5ApplicationMenu::buttonClicked()
{
    QPushButton *p = ::qobject_cast<QPushButton *>(sender());
    if (p) {
        m_selected = m_actions.value(p).data();
        if (m_selected)
            accept();
    }
}


QAction *QMaemo5ApplicationMenu::selectedAction() const
{
    return m_selected;
}


#endif //Q_WS_MAEMO_5

QT_END_NAMESPACE

